# -*- coding: utf-8 -*-
# Generated by Django 1.10 on 2017-01-03 16:11
from __future__ import unicode_literals

from django.db import migrations
import elasticsearch_django.db.fields


class Migration(migrations.Migration):

    dependencies = [
        ('elasticsearch_django', '0004_auto_20161129_1135'),
    ]

    operations = [
        migrations.AlterField(
            model_name='searchquery',
            name='hits',
            field=elasticsearch_django.db.fields.JSONField(help_text=b'The list of meta info for each of the query matches returned.'),
        ),
        migrations.AlterField(
            model_name='searchquery',
            name='query',
            field=elasticsearch_django.db.fields.JSONField(help_text=b'The raw ElasticSearch DSL query.'),
        ),
    ]
